#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from connect.client.exceptions import ClientError  # noqa
from connect.client.fluent import ConnectClient  # noqa
from connect.client.rql import R  # noqa



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from connect.client.models.mixins import (
    ActionMixin,
    AsyncActionMixin,
    AsyncCollectionMixin,
    AsyncResourceMixin,
    CollectionMixin,
    ResourceMixin,
)
from connect.client.models.resourceset import AsyncResourceSet, ResourceSet
from connect.client.rql import R


class _NSBase:
    """
    A namespace is a group of related collections.
    """

    def __init__(self, client, path):
        """
        Create a new NS instance.

        :param client: the client instance
        :type client: ConnectClient
        :param path: path name of the namespace
        :type path: str
        """
        self._client = client
        self._path = path

    @property
    def path(self):
        return self._path

    def __getattr__(self, name):
        """
        Returns a collection object by its name.

        :param name: the name of the Collection object.
        :type name: str
        :return: The Collection named ``name``.
        :rtype: Collection
        """
        if '_' in name:
            name = name.replace('_', '-')
        return self.collection(name)

    def __iter__(self):
        raise TypeError('A Namespace object is not iterable.')

    def __call__(self, name):
        return self.ns(name)

    def collection(self, name):
        """
        Returns the collection called ``name``.

        :param name: The name of the collection.
        :type name: str
        :raises TypeError: if the ``name`` is not a string.
        :raises ValueError: if the ``name`` is blank.
        :raises NotFoundError: if the ``name`` does not exist.
        :return: The collection called ``name``.
        :rtype: Collection
        """
        if not isinstance(name, str):
            raise TypeError('`name` must be a string.')

        if not name:
            raise ValueError('`name` must not be blank.')

        return self._get_collection_class()(
            self._client,
            f'{self._path}/{name}',
        )

    def ns(self, name):
        """
        Returns the namespace called ``name``.

        :param name: The name of the namespace.
        :type name: str
        :raises TypeError: if the ``name`` is not a string.
        :raises ValueError: if the ``name`` is blank.
        :return: The namespace called ``name``.
        :rtype: NS
        """
        if not isinstance(name, str):
            raise TypeError('`name` must be a string.')

        if not name:
            raise ValueError('`name` must not be blank.')

        return self._get_namespace_class()(
            self._client,
            f'{self._path}/{name}',
        )

    def help(self):
        """
        Output the namespace documentation to the console.

        :return: self
        :rtype: NS
        """
        self._client.print_help(self)
        return self

    def _get_collection_class(self):
        raise NotImplementedError()

    def _get_namespace_class(self):
        raise NotImplementedError()


class NS(_NSBase):
    def _get_collection_class(self):
        return Collection

    def _get_namespace_class(self):
        return NS


class AsyncNS(_NSBase):
    def _get_collection_class(self):
        return AsyncCollection

    def _get_namespace_class(self):
        return AsyncNS


class _CollectionBase:
    """
    A collection is a group of operations on a resource.
    """
    def __init__(self, client, path):
        """
        Create a new Collection instance.

        :param client: the client instance
        :type client: ConnectClient
        :param path: path name of the collection
        :type path: str
        """
        self._client = client
        self._path = path

    @property
    def path(self):
        return self._path

    def __iter__(self):
        raise TypeError('A Collection object is not iterable.')

    def __getitem__(self, resource_id):
        """
        Return a Resource object representing the resource
        identified by ``resource_id``.

        :param resource_id: The identifier of the resource
        :type resource_id: str, int
        :return: the Resource instance identified by ``resource_id``.
        :rtype: Resource
        """
        return self.resource(resource_id)

    def all(self):
        """
        Return a ResourceSet instance.

        :return: a ResourceSet instance.
        :rtype: ResourceSet
        """
        return self._get_resourceset_class()(
            self._client,
            self._path,
        )

    def filter(self, *args, **kwargs):
        """
        Returns a ResourceSet object.
        The returned ResourceSet object will be filtered based on
        the arguments and keyword arguments.

        Arguments can be RQL filter expressions as strings
        or R objects.

        Ex.

        .. code-block:: python

            rs = collection.filter('eq(field,value)', 'eq(another.field,value2)')
            rs = collection.filter(R().field.eq('value'), R().another.field.eq('value2'))

        All the arguments will be combined with logical ``and``.

        Filters can be also specified as keyword argument using the ``__`` (double underscore)
        notation.

        Ex.

        .. code-block:: python

            rs = collection.filter(
                field=value,
                another__field=value,
                field2__in=('a', 'b'),
                field3__null=True,
            )

        Also keyword arguments will be combined with logical ``and``.


        :raises TypeError: If arguments are neither strings nor R objects.
        :return: A ResourceSet with the filters applied.
        :rtype: ResourceSet
        """
        query = R()
        for arg in args:
            if isinstance(arg, str):
                query &= R(_expr=arg)
                continue
            if isinstance(arg, R):
                query &= arg
                continue
            raise TypeError(f'arguments must be string or R not {type(arg)}')

        if kwargs:
            query &= R(**kwargs)

        return self._get_resourceset_class()(
            self._client,
            self._path,
            query=query,
        )

    def resource(self, resource_id):
        """
        Returns an Resource object.

        :param resource_id: The resource identifier.
        :type resource_id: str, int
        :return: The Resource identified by ``resource_id``.
        :rtype: Resource
        """
        if not isinstance(resource_id, (str, int)):
            raise TypeError('`resource_id` must be a string or int.')

        if not resource_id:
            raise ValueError('`resource_id` must not be blank.')

        return self._get_resource_class()(
            self._client,
            f'{self._path}/{resource_id}',
        )

    def help(self):
        """
        Output the collection documentation to the console.

        :return: self
        :rtype: Collection
        """
        self._client.print_help(self)
        return self

    def _get_resource_class(self):
        return NotImplementedError()

    def _get_resourceset_class(self):
        return NotImplementedError()


class Collection(_CollectionBase, CollectionMixin):
    def _get_resource_class(self):
        return Resource

    def _get_resourceset_class(self):
        return ResourceSet


class AsyncCollection(_CollectionBase, AsyncCollectionMixin):
    def _get_resource_class(self):
        return AsyncResource

    def _get_resourceset_class(self):
        return AsyncResourceSet


class _ResourceBase:
    """Represent a generic resource."""
    def __init__(self, client, path):
        """
        Create a new Resource instance.

        :param client: the client instance
        :type client: ConnectClient
        :param path: path name of the resource
        :type path: str
        :param specs: OpenAPI specs, defaults to None
        :type specs: ResourceInfo, optional
        """
        self._client = client
        self._path = path

    @property
    def path(self):
        return self._path

    def __getattr__(self, name):
        """
        Returns a nested Collection object called ``name``.

        :param name: The name of the Collection to retrieve.
        :type name: str
        :return: a Collection called ``name``.
        :rtype: Collection
        """
        if '_' in name:
            name = name.replace('_', '-')
        return self.collection(name)

    def __call__(self, name):
        return self.action(name)

    def collection(self, name):
        """
        Returns the collection called ``name``.

        :param name: The name of the collection.
        :type name: str
        :raises TypeError: if the ``name`` is not a string.
        :raises ValueError: if the ``name`` is blank.
        :raises NotFoundError: if the ``name`` does not exist.
        :return: The collection called ``name``.
        :rtype: Collection
        """
        if not isinstance(name, str):
            raise TypeError('`name` must be a string.')

        if not name:
            raise ValueError('`name` must not be blank.')

        return self._get_collection_class()(
            self._client,
            f'{self._path}/{name}',
        )

    def action(self, name):
        """
        Returns the action called ``name``.

        :param name: The name of the action.
        :type name: str
        :raises TypeError: if the ``name`` is not a string.
        :raises ValueError: if the ``name`` is blank.
        :raises NotFoundError: if the ``name`` does not exist.
        :return: The action called ``name``.
        :rtype: Action
        """
        if not isinstance(name, str):
            raise TypeError('`name` must be a string.')

        if not name:
            raise ValueError('`name` must not be blank.')

        return self._get_action_class()(
            self._client,
            f'{self._path}/{name}',
        )

    def help(self):
        """
        Output the resource documentation to the console.

        :return: self
        :rtype: Resource
        """
        self._client.print_help(self)
        return self

    def _get_collection_class(self):
        raise NotImplementedError()

    def _get_action_class(self):
        raise NotImplementedError()


class Resource(_ResourceBase, ResourceMixin):
    def _get_collection_class(self):
        return Collection

    def _get_action_class(self):
        return Action


class AsyncResource(_ResourceBase, AsyncResourceMixin):
    def _get_collection_class(self):
        return AsyncCollection

    def _get_action_class(self):
        return AsyncAction


class _ActionBase:
    """
    This class represent an action that can be executed on a resource.
    """
    def __init__(self, client, path):
        """
        Create a new Action instance.

        :param client: the client instance
        :type client: ConnectClient
        :param path: path name of the action
        :type path: str
        :param specs: OpenAPI specs, defaults to None
        :type specs: ActionInfo, optional
        """
        self._client = client
        self._path = path

    @property
    def path(self):
        return self._path

    def help(self):
        """
        Output the action documentation to the console.

        :return: self
        :rtype: Action
        """
        self._client.print_help(self)
        return self


class Action(_ActionBase, ActionMixin):
    pass


class AsyncAction(_ActionBase, AsyncActionMixin):
    pass



# -*- coding: utf-8 -*-

#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#

from pkg_resources import DistributionNotFound, get_distribution


def get_version():
    try:
        return get_distribution('connect-openapi-client').version
    except DistributionNotFound:
        return '0.0.0'



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from io import StringIO
from functools import partial

import requests

import yaml


class OpenAPISpecs:

    def __init__(self, location):
        self._location = location
        self._specs = self._load()

    @property
    def title(self):
        return self._specs['info']['title'] if self._specs else None

    @property
    def description(self):
        return self._specs['info']['description'] if self._specs else None

    @property
    def version(self):
        return self._specs['info']['version'] if self._specs else None

    @property
    def tags(self):
        return self._specs.get('tags') if self._specs else None

    def exists(self, method, path):
        p = self._get_path(path)
        if not p:
            return False
        info = self._specs['paths'][p]
        return method.lower() in info

    def get_namespaces(self):
        def _is_namespace(path):
            comp = path[1:].split('/', 1)
            return len(comp) > 1 and not comp[1].startswith('{')
        return sorted(
            list(
                {
                    p[1:].split('/', 1)[0]
                    for p in filter(_is_namespace, self._specs['paths'].keys())
                },
            ),
        )

    def get_collections(self):
        namespaces = self.get_namespaces()
        cols = set()
        for p in self._specs['paths'].keys():
            name = p[1:].split('/', 1)[0]
            if name not in namespaces:
                cols.add(name)

        return sorted(cols)

    def get_namespaced_collections(self, path):
        nested = filter(lambda x: x[1:].startswith(path), self._specs['paths'].keys())
        collections = set()
        for p in nested:
            splitted = p[len(path) + 1:].split('/', 2)
            if self._is_collection(p) and len(splitted) == 2:
                collections.add(splitted[1])
        return list(sorted(collections))

    def get_collection(self, path):
        return self._get_info(path)

    def get_resource(self, path):
        return self._get_info(path)

    def get_action(self, path):
        return self._get_info(path)

    def get_actions(self, path):
        p = self._get_path(path)
        nested = filter(
            lambda x: x.startswith(p) and x != p,
            self._specs['paths'].keys(),
        )

        actions = set()
        descriptions = {}
        for np in nested:
            name = np[len(p) + 1:]
            actions.add(name)
            info = self._specs['paths'][np]
            summary = info['summary'] if 'summary' in info else ''
            if summary:
                descriptions[name] = summary
        return [
            (name, descriptions.get(name))
            for name in sorted(actions)
        ]

    def get_nested_namespaces(self, path):
        def _is_nested_namespace(base_path, path):
            if path[1:].startswith(base_path):
                comp = path[1:].split('/')
                return (
                    len(comp) > 1
                    and not comp[-1].startswith('{')
                )
            return False

        nested = filter(
            partial(_is_nested_namespace, path),
            self._specs['paths'].keys(),
        )
        current_level = len(path[1:].split('/'))
        nested_namespaces = []
        for ns in nested:
            name = ns[1:].split('/')[current_level]
            if not self._is_collection(f'/{path}/{name}'):
                nested_namespaces.append(name)
        return nested_namespaces

    def get_nested_collections(self, path):
        p = self._get_path(path)
        nested = filter(
            lambda x: x.startswith(p[0:p.rindex('{')]) and x != p,
            self._specs['paths'].keys(),
        )
        cut_pos = p.count('/')
        collections = set()
        descriptions = {}
        for np in nested:
            splitted = np[1:].split('/')
            name = splitted[cut_pos]
            info = self._specs['paths'][np]
            method_info = info['get'] if 'get' in info else info['post']
            operation_id = method_info['operationId']
            if self._is_action(operation_id):
                continue
            collections.add(name)
            summary = info['summary'] if 'summary' in info else ''
            if summary:
                descriptions[name] = summary
        return [
            (name, descriptions.get(name))
            for name in sorted(collections)
        ]

    def _load(self):
        if self._location.startswith('http'):
            return self._load_from_url()
        return self._load_from_fs()

    def _load_from_url(self):
        res = requests.get(self._location, stream=True)
        if res.status_code == 200:
            result = StringIO()
            for chunk in res.iter_content(chunk_size=8192):
                result.write(str(chunk, encoding='utf-8'))
            result.seek(0)
            return yaml.safe_load(result)
        res.raise_for_status()

    def _load_from_fs(self):
        with open(self._location, 'r') as f:
            return yaml.safe_load(f)

    def _get_path(self, path):
        if '?' in path:
            path, _ = path.split('?', 1)
        components = path.split('/')
        for p in self._specs['paths'].keys():
            cmps = p[1:].split('/')
            if len(cmps) != len(components):
                continue
            for idx, comp in enumerate(components):
                if cmps[idx].startswith('{'):
                    continue
                if cmps[idx] != comp:
                    break
            else:
                return p

    def _get_info(self, path):
        p = self._get_path(path)
        return self._specs['paths'][p] if p else None

    def _is_action(self, operation_id):
        op_id_cmps = operation_id.rsplit('_', 2)
        return op_id_cmps[-2] not in ('list', 'retrieve')

    def _is_collection(self, path):
        path_length = len(path[1:].split('/'))
        for p in self._specs['paths'].keys():
            comp = p[1:].split('/')
            if not p.startswith(path):
                continue
            if p == path:
                return True
            if len(comp) > path_length and comp[path_length].startswith('{'):
                return True
        return False



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import copy

from connect.client.models.exceptions import NotYetEvaluatedError
from connect.client.models.iterators import (
    AsyncResourceIterator,
    AsyncValuesListIterator,
    ResourceIterator,
    ValuesListIterator,
    aiter,
)
from connect.client.utils import parse_content_range, resolve_attribute
from connect.client.rql import R


class _ResourceSetBase:
    """
    Represent a set of resources.
    """
    def __init__(
        self,
        client,
        path,
        query=None,
    ):

        self._client = client
        self._path = path
        self._query = query or R()
        self._results = None
        self._limit = self._client.default_limit or 100
        self._offset = 0
        self._slice = False
        self._content_range = None
        self._fields = None
        self._search = None
        self._select = []
        self._ordering = []
        self._config = {}

    @property
    def path(self):
        return self._path

    @property
    def query(self):
        return self._query

    @property
    def content_range(self):
        return self._content_range

    def configure(self, **kwargs):
        """
        Set the keyword arguments that needs to be forwarded to
        the underlying ``requests`` call.

        :return: This ResourceSet object.
        :rtype: ResourceSet
        """
        copy = self._copy()
        copy._config = kwargs or {}
        return copy

    def limit(self, limit):
        """
        Set the number of results to be fetched from the remote
        endpoint at once.

        :param limit: maximum number of results to fetch in a batch.
        :type limit: int
        :raises TypeError: if `limit` is not an integer.
        :raises ValueError: if `limit` is not positive non-zero.
        :return: A copy of this ResourceSet class with the new limit.
        :rtype: ResourceSet
        """
        if not isinstance(limit, int):
            raise TypeError('`limit` must be an integer.')

        if limit <= 0:
            raise ValueError('`limit` must be a positive, non-zero integer.')

        copy = self._copy()
        copy._limit = limit
        return copy

    def order_by(self, *fields):
        """
        Add fields for ordering.

        :return: This ResourceSet object.
        :rtype: ResourceSet
        """
        copy = self._copy()
        copy._ordering.extend(fields)
        return copy

    def select(self, *fields):
        """
        Apply the RQL ``select`` operator to
        this ResourceSet object.

        :return: This ResourceSet object.
        :rtype: ResourceSet
        """
        copy = self._copy()
        copy._select.extend(fields)
        return copy

    def filter(self, *args, **kwargs):
        """
        Applies filters to this ResourceSet object.

        Arguments can be RQL filter expressions as strings
        or R objects.

        Ex.

        .. code-block:: python

            rs = rs.filter('eq(field,value)', 'eq(another.field,value2)')
            rs = rs.filter(R().field.eq('value'), R().another.field.eq('value2'))

        All the arguments will be combined with logical ``and``.

        Filters can be also specified as keyword argument using the ``__`` (double underscore)
        notation.

        Ex.

        .. code-block:: python

            rs = rs.filter(
                field=value,
                another__field=value,
                field2__in=('a', 'b'),
                field3__null=True,
            )

        Also keyword arguments will be combined with logical ``and``.


        :raises TypeError: If arguments are neither strings nor R objects.
        :return: This ResourceSet object.
        :rtype: ResourceSet
        """
        copy = self._copy()
        for arg in args:
            if isinstance(arg, str):
                copy._query &= R(_expr=arg)
                continue
            if isinstance(arg, R):
                copy._query &= arg
                continue
            raise TypeError(f'arguments must be string or R not {type(arg)}')

        if kwargs:
            copy._query &= R(**kwargs)

        return copy

    def all(self):
        """
        Returns a copy of the current ResourceSet.

        :return: A copy of this ResourceSet.
        :rtype: ResourceSet
        """
        return self._copy()

    def search(self, term):
        """
        Create a copy of the current ResourceSet with
        the search set to `term`.

        :param term: The term to search for.
        :type term: str
        :return: A copy of the current ResourceSet.
        :rtype: ResourceSet
        """
        copy = self._copy()
        copy._search = term
        return copy

    def values_list(self, *fields):
        """
        Returns a flat dictionary containing only the fields passed as arguments
        for each resource that belongs to this ResourceSet.

        Nested field can be specified using dot notation.

        Ex.

        .. code-block:: python

        values = rs.values_list('field', 'nested.field')

        :return: A list of dictionaries containing field,value pairs.
        :rtype: list
        """
        if self._results:
            self._fields = fields
            return [
                self._get_values(item)
                for item in self._results
            ]

        copy = self._copy()
        copy._fields = fields
        return copy

    def _get_values(self, item):
        return {
            field: resolve_attribute(field, item)
            for field in self._fields
        }

    def _build_qs(self):
        qs = ''
        if self._select:
            qs += f'&select({",".join(self._select)})'
        if self._query:
            qs += f'&{str(self._query)}'
        if self._ordering:
            qs += f'&ordering({",".join(self._ordering)})'
        return qs[1:] if qs else ''

    def _get_request_url(self):
        url = f'{self._path}'
        qs = self._build_qs()
        if qs:
            url = f'{url}?{qs}'

        return url

    def _get_request_kwargs(self):
        config = copy.deepcopy(self._config)
        config.setdefault('params', {})

        config['params'].update({
            'limit': self._limit,
            'offset': self._offset,
        })

        if self._search:
            config['params']['search'] = self._search

        return config

    def _copy(self):
        rs = self.__class__(self._client, self._path, self._query)
        rs._limit = self._limit
        rs._offset = self._offset
        rs._slice = self._slice
        rs._fields = self._fields
        rs._search = self._search
        rs._select = copy.copy(self._select)
        rs._ordering = copy.copy(self._ordering)
        rs._config = copy.deepcopy(self._config)

        return rs

    def help(self):
        """
        Output the ResourceSet documentation to the console.

        :return: self
        :rtype: ResourceSet
        """
        self._client.print_help(self)
        return self


class ResourceSet(_ResourceSetBase):

    def __iter__(self):
        """
        Returns an iterator to iterate over the set of resources.

        :return: A resources iterator.
        :rtype: ResourceSet
        """
        if self._results is None:
            return self._iterator()
        return iter(self._results)

    def __bool__(self):
        """
        Return True if the ResourceSet contains at least a resource
        otherwise return False.

        :return: True if contains a resource otherwise False.
        :rtype: bool
        """
        self._fetch_all()
        return bool(self._results)

    def __getitem__(self, key):  # noqa: CCR001
        """
        If called with and integer index, returns the item
        at index ``key``.

        If key is a slice, set the pagination limit and offset
        accordingly.

        :param key: index or slice.
        :type key: int, slice
        :raises TypeError: If ``key`` is neither an integer nor a slice.
        :return: The resource at index ``key`` or self if ``key`` is a slice.
        :rtype: dict, ResultSet
        """
        if not isinstance(key, (int, slice)):
            raise TypeError('ResourceSet indices must be integers or slices.')

        if isinstance(key, slice) and (key.start is None or key.stop is None):
            raise ValueError('Both start and stop indexes must be specified.')

        if (not isinstance(key, slice) and (key < 0)) or (
            isinstance(key, slice) and (key.start < 0 or key.stop < 0)
        ):
            raise ValueError('Negative indexing is not supported.')

        if isinstance(key, slice) and not (key.step is None or key.step == 0):
            raise ValueError('Indexing with step is not supported.')

        if self._results is not None:
            return self._results[key]

        if isinstance(key, int):
            copy = self._copy()
            copy._limit = 1
            copy._offset = key
            copy._fetch_all()
            return copy._results[0] if copy._results else None

        copy = self._copy()
        copy._offset = key.start
        copy._limit = key.stop - key.start
        copy._slice = True
        return copy

    def count(self):
        """
        Returns the total number of resources within this ResourceSet object.

        :return: The total number of resources present.
        :rtype: int
        """
        if not self._content_range:
            url = self._get_request_url()
            kwargs = self._get_request_kwargs()
            kwargs['params']['limit'] = 0
            self._execute_request(url, kwargs)
        return self._content_range.count

    def first(self):
        """
        Returns the first resource that belongs to this ResourceSet object
        or None if the ResourceSet doesn't contains resources.

        :return: The first resource.
        :rtype: dict, None
        """
        self._fetch_all()
        return self._results[0] if self._results else None

    def _iterator(self):
        args = (
            self,
            self._client,
            self._path,
            self._build_qs(),
            self._get_request_kwargs(),
        )
        iterator = (
            ValuesListIterator(*args, fields=self._fields)
            if self._fields else ResourceIterator(*args)
        )
        return iterator

    def _execute_request(self, url, kwargs):
        results = self._client.get(url, **kwargs)
        self._content_range = parse_content_range(
            self._client.response.headers.get('Content-Range'),
        )
        return results

    def _fetch_all(self):
        if self._results is None:
            self._results = self._execute_request(
                self._get_request_url(),
                self._get_request_kwargs(),
            )


class AsyncResourceSet(_ResourceSetBase):

    def __aiter__(self):
        """
        Returns an iterator to iterate over the set of resources.

        :return: A resources iterator.
        :rtype: ResourceSet
        """
        if self._results is None:
            return self._iterator()
        return aiter(self._results)

    def __bool__(self):
        """
        Return True if the ResourceSet contains at least a resource
        otherwise return False.

        :return: True if contains a resource otherwise False.
        :rtype: bool
        """
        if self._results is None:
            raise NotYetEvaluatedError()
        return bool(self._results)

    def __getitem__(self, key):  # noqa: CCR001
        """
        If called with and integer index, returns the item
        at index ``key``.

        If key is a slice, set the pagination limit and offset
        accordingly.

        :param key: index or slice.
        :type key: int, slice
        :raises TypeError: If ``key`` is neither an integer nor a slice.
        :return: The resource at index ``key`` or self if ``key`` is a slice.
        :rtype: dict, ResultSet
        """
        if not isinstance(key, (int, slice)):
            raise TypeError('ResourceSet indices must be integers or slices.')

        if isinstance(key, slice) and (key.start is None or key.stop is None):
            raise ValueError('Both start and stop indexes must be specified.')

        if (not isinstance(key, slice) and (key < 0)) or (
            isinstance(key, slice) and (key.start < 0 or key.stop < 0)
        ):
            raise ValueError('Negative indexing is not supported.')

        if isinstance(key, slice) and not (key.step is None or key.step == 0):
            raise ValueError('Indexing with step is not supported.')

        if self._results is not None:
            return self._results[key]

        if isinstance(key, int):
            raise NotYetEvaluatedError()

        copy = self._copy()
        copy._offset = key.start
        copy._limit = key.stop - key.start
        copy._slice = True
        return copy

    async def count(self):
        """
        Returns the total number of resources within this ResourceSet object.

        :return: The total number of resources present.
        :rtype: int
        """
        if not self._content_range:
            url = self._get_request_url()
            kwargs = self._get_request_kwargs()
            kwargs['params']['limit'] = 0
            await self._execute_request(url, kwargs)
        return self._content_range.count

    async def first(self):
        """
        Returns the first resource that belongs to this ResourceSet object
        or None if the ResourceSet doesn't contains resources.

        :return: The first resource.
        :rtype: dict, None
        """
        await self._fetch_all()
        return self._results[0] if self._results else None

    def _iterator(self):
        args = (
            self,
            self._client,
            self._path,
            self._build_qs(),
            self._get_request_kwargs(),
        )

        iterator = (
            AsyncValuesListIterator(*args, fields=self._fields)
            if self._fields else AsyncResourceIterator(*args)
        )
        return iterator

    async def _execute_request(self, url, kwargs):
        results = await self._client.get(url, **kwargs)
        self._content_range = parse_content_range(
            self._client.response.headers.get('Content-Range'),
        )
        return results

    async def _fetch_all(self):
        if self._results is None:
            self._results = await self._execute_request(
                self._get_request_url(),
                self._get_request_kwargs(),
            )



import time

from locust import HttpUser

from requests import RequestException

from connect.client import ConnectClient


class _LocustConnectClient(ConnectClient):
    def __init__(self, base_url, request_event, user, *args, **kwargs):
        self.base_url = base_url
        self.request_event = request_event
        self.user = user

        super().__init__(
            kwargs['connect_api_key'],
            self.base_url,
            use_specs=False,
            max_retries=0,
            timeout=(5, 120),
        )

    def _execute_http_call(self, method, url, kwargs):
        start_time = time.perf_counter()
        exc = None
        try:
            super()._execute_http_call(method, url, kwargs)
        except RequestException as e:
            exc = e

        response = self.response.history and self.response.history[0] or self.response
        request_meta = {
            'request_type': method,
            'response_time': (time.perf_counter() - start_time) * 1000,
            'name': response.request.path_url,
            'context': {},
            'response': self.response,
            'exception': exc,
            'response_length': len(self.response.content or b''),
        }

        self.request_event.fire(**request_meta)


class ConnectHttpUser(HttpUser):
    abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.client = _LocustConnectClient(
            self.host,
            request_event=self.environment.events.request,
            user=self,
            connect_api_key=self.environment.parsed_options.connect_api_key,
        )



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import threading
from json.decoder import JSONDecodeError

from connect.client.constants import CONNECT_ENDPOINT_URL, CONNECT_SPECS_URL
from connect.client.mixins import AsyncClientMixin, SyncClientMixin
from connect.client.models import AsyncCollection, AsyncNS, Collection, NS
from connect.client.utils import get_headers
from connect.client.help_formatter import DefaultFormatter
from connect.client.openapi import OpenAPISpecs


class _ConnectClientBase(threading.local):

    """
    Connect ReST API client.
    """
    def __init__(
        self,
        api_key,
        endpoint=None,
        use_specs=True,
        specs_location=None,
        validate_using_specs=True,
        default_headers=None,
        default_limit=100,
        max_retries=0,
        logger=None,
        timeout=(180.0, 180.0),
    ):
        """
        Create a new instance of the ConnectClient.

        :param api_key: The API key used for authentication.
        :type api_key: str
        :param endpoint: The API endpoint, defaults to CONNECT_ENDPOINT_URL
        :type endpoint: str, optional
        :param specs_location: The Connect OpenAPI specification local path or URL, defaults to
                               CONNECT_SPECS_URL
        :type specs_location: str, optional
        :param default_headers: Http headers to apply to each request, defaults to {}
        :type default_headers: dict, optional
        :param logger: HTTPP Request logger class, defaults to None
        :type logger: RequestLogger, optional
        """
        if default_headers and 'Authorization' in default_headers:
            raise ValueError('`default_headers` cannot contains `Authorization`')

        self.endpoint = endpoint or CONNECT_ENDPOINT_URL
        self.api_key = api_key
        self.default_headers = default_headers or {}
        self.default_limit = default_limit
        self.max_retries = max_retries
        self._use_specs = use_specs
        self._validate_using_specs = validate_using_specs
        self.specs_location = specs_location or CONNECT_SPECS_URL
        self.specs = None
        if self._use_specs:
            self.specs = OpenAPISpecs(self.specs_location)
        self.response = None
        self.logger = logger
        self._help_formatter = DefaultFormatter(self.specs)
        self.timeout = timeout

    def __getattr__(self, name):
        """
        Returns a collection object called ``name``.

        :param name: The name of the collection to retrieve.
        :type name: str
        :return: a collection called ``name``.
        :rtype: Collection
        """
        if '_' in name:
            name = name.replace('_', '-')
        return self.collection(name)

    def __call__(self, name):
        return self.ns(name)

    def ns(self, name):
        """
        Returns the namespace called ``name``.

        :param name: The name of the namespace to access.
        :type name: str
        :return: The namespace called ``name``.
        :rtype: NS
        """
        if not isinstance(name, str):
            raise TypeError('`name` must be a string.')

        if not name:
            raise ValueError('`name` must not be blank.')

        return self._get_namespace_class()(self, name)

    def collection(self, name):
        """
        Returns the collection called ``name``.

        :param name: The name of the collection to access.
        :type name: str
        :return: The collection called ``name``.
        :rtype: Collection
        """
        if not isinstance(name, str):
            raise TypeError('`name` must be a string.')

        if not name:
            raise ValueError('`name` must not be blank.')

        return self._get_collection_class()(
            self,
            name,
        )

    def print_help(self, obj):
        print()
        print(self._help_formatter.format(obj))

    def help(self):
        self.print_help(None)
        return self

    def _get_collection_class(self):
        raise NotImplementedError()

    def _get_namespace_class(self):
        raise NotImplementedError()

    def _prepare_call_kwargs(self, kwargs):
        kwargs = kwargs or {}
        if 'headers' in kwargs:
            kwargs['headers'].update(get_headers(self.api_key))
        else:
            kwargs['headers'] = get_headers(self.api_key)
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout
        if self.default_headers:
            kwargs['headers'].update(self.default_headers)
        return kwargs

    def _get_api_error_details(self):
        if self.response is not None:
            try:
                error = self.response.json()
                if 'error_code' in error and 'errors' in error:
                    return error
            except JSONDecodeError:
                pass


class ConnectClient(_ConnectClientBase, SyncClientMixin):
    def _get_collection_class(self):
        return Collection

    def _get_namespace_class(self):
        return NS


class AsyncConnectClient(_ConnectClientBase, AsyncClientMixin):
    def _get_collection_class(self):
        return AsyncCollection

    def _get_namespace_class(self):
        return AsyncNS



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import platform
from collections import namedtuple

from connect.client.version import get_version

ContentRange = namedtuple('ContentRange', ('first', 'last', 'count'))


def _get_user_agent():
    version = get_version()
    pimpl = platform.python_implementation()
    pver = platform.python_version()
    sysname = platform.system()
    sysver = platform.release()
    ua = f'connect-fluent/{version} {pimpl}/{pver} {sysname}/{sysver}'
    return {'User-Agent': ua}


def get_headers(api_key):
    headers = {'Authorization': api_key}
    headers.update(_get_user_agent())
    return headers


def parse_content_range(value):
    if not value:
        return
    _, info = value.split()
    first_last, count = info.split('/')
    first, last = first_last.split('-')
    return ContentRange(int(first), int(last), int(count))


def resolve_attribute(attr, data):
    try:
        for comp in attr.split('.'):
            data = data.get(comp)
        return data
    except:  # noqa
        pass


def get_values(item, fields):
    return {field: resolve_attribute(field, item) for field in fields}



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import time

import httpx

import requests

from httpx import HTTPError

from requests.exceptions import RequestException, Timeout

from connect.client.exceptions import ClientError


class SyncClientMixin:

    def get(self, url, **kwargs):
        return self.execute('get', url, **kwargs)

    def create(self, url, payload=None, **kwargs):
        kwargs = kwargs or {}

        if payload:
            kwargs['json'] = payload

        return self.execute('post', url, **kwargs)

    def update(self, url, payload=None, **kwargs):
        kwargs = kwargs or {}

        if payload:
            kwargs['json'] = payload

        return self.execute('put', url, **kwargs)

    def delete(self, url, **kwargs):
        return self.execute('delete', url, **kwargs)

    def execute(self, method, path, **kwargs):
        if (
            self._use_specs
            and self._validate_using_specs
            and not self.specs.exists(method, path)
        ):
            # TODO more info, specs version, method etc
            raise ClientError(f'The path `{path}` does not exist.')

        url = f'{self.endpoint}/{path}'

        kwargs = self._prepare_call_kwargs(kwargs)

        self.response = None

        try:
            self._execute_http_call(method, url, kwargs)

            if self.response.status_code == 204:
                return None
            if self.response.headers.get('Content-Type', '').startswith('application/json'):
                return self.response.json()
            else:
                return self.response.content

        except RequestException as re:
            api_error = self._get_api_error_details() or {}
            status_code = self.response.status_code if self.response is not None else None
            raise ClientError(status_code=status_code, **api_error) from re

    def _execute_http_call(self, method, url, kwargs):  # noqa: CCR001
        retry_count = 0
        while True:
            if self.logger:
                self.logger.log_request(method, url, kwargs)
            try:
                self.response = requests.request(method, url, **kwargs)
                if self.logger:
                    self.logger.log_response(self.response)
            except Timeout:
                if retry_count < self.max_retries:
                    retry_count += 1
                    time.sleep(1)
                    continue
                raise

            if (  # pragma: no branch
                self.response.status_code >= 500
                and retry_count < self.max_retries
            ):
                retry_count += 1
                time.sleep(1)
                continue
            break  # pragma: no cover
        if self.response.status_code >= 400:
            self.response.raise_for_status()


class AsyncClientMixin:

    async def get(self, url, **kwargs):
        return await self.execute('get', url, **kwargs)

    async def create(self, url, payload=None, **kwargs):
        kwargs = kwargs or {}

        if payload:
            kwargs['json'] = payload

        return await self.execute('post', url, **kwargs)

    async def update(self, url, payload=None, **kwargs):
        kwargs = kwargs or {}

        if payload:
            kwargs['json'] = payload

        return await self.execute('put', url, **kwargs)

    async def delete(self, url, **kwargs):
        return await self.execute('delete', url, **kwargs)

    async def execute(self, method, path, **kwargs):
        if (
            self._use_specs
            and self._validate_using_specs
            and not self.specs.exists(method, path)
        ):
            # TODO more info, specs version, method etc
            raise ClientError(f'The path `{path}` does not exist.')

        url = f'{self.endpoint}/{path}'

        kwargs = self._prepare_call_kwargs(kwargs)

        url, kwargs = self._fix_url_params(url, kwargs)

        self.response = None

        try:
            await self._execute_http_call(method, url, kwargs)
            if self.response.status_code == 204:
                return None
            if self.response.headers.get('Content-Type', '').startswith('application/json'):
                return self.response.json()
            else:
                return self.response.content

        except HTTPError as re:
            api_error = self._get_api_error_details() or {}
            status_code = self.response.status_code if self.response is not None else None
            raise ClientError(status_code=status_code, **api_error) from re

    async def _execute_http_call(self, method, url, kwargs):
        retry_count = 0
        while True:
            if self.logger:
                self.logger.log_request(method, url, kwargs)

            async with httpx.AsyncClient() as client:
                self.response = await client.request(method, url, **kwargs)

            if self.logger:
                self.logger.log_response(self.response)

            if (  # pragma: no branch
                self.response.status_code >= 500
                and retry_count < self.max_retries
            ):
                retry_count += 1
                time.sleep(1)
                continue
            break  # pragma: no cover
        if self.response.status_code >= 400:
            self.response.raise_for_status()

    def _fix_url_params(self, url, kwargs):
        if 'params' in kwargs:
            params = kwargs.pop('params')
            qs_fragment = '&'.join([f'{k}={v}' for k, v in params.items()])
            join = '?' if '?' not in url else '&'
            url = f'{url}{join}{qs_fragment}'
        return url, kwargs



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from datetime import date, datetime
from decimal import Decimal


COMP = ('eq', 'ne', 'lt', 'le', 'gt', 'ge')
SEARCH = ('like', 'ilike')
LIST = ('in', 'out')
NULL = 'null'
EMPTY = 'empty'

KEYWORDS = (*COMP, *SEARCH, *LIST, NULL, EMPTY)


def parse_kwargs(query_dict):
    query = []
    for lookup, value in query_dict.items():
        tokens = lookup.split('__')
        if len(tokens) == 1:
            #  field=value
            field = tokens[0]
            value = to_rql_value('eq', value)
            query.append(f'eq({field},{value})')
            continue
        op = tokens[-1]
        if op not in KEYWORDS:
            # field__nested=value
            field = '.'.join(tokens)
            value = to_rql_value('eq', value)
            query.append(f'eq({field},{value})')
            continue
        field = '.'.join(tokens[:-1])
        if op in COMP or op in SEARCH:
            value = to_rql_value(op, value)
            query.append(f'{op}({field},{value})')
            continue
        if op in LIST:
            value = to_rql_value(op, value)
            query.append(f'{op}({field},({value}))')
            continue

        cmpop = 'eq' if value is True else 'ne'
        expr = 'null()' if op == NULL else 'empty()'
        query.append(f'{cmpop}({field},{expr})')

    return query


def to_rql_value(op, value):
    if op not in LIST:
        if isinstance(value, str):
            return value
        if isinstance(value, bool):
            return 'true' if value else 'false'
        if isinstance(value, (int, float, Decimal)):
            return str(value)
        if isinstance(value, (date, datetime)):
            return value.isoformat()
    if op in LIST and isinstance(value, (list, tuple)):
        return ','.join(value)
    raise TypeError(f"the `{op}` operator doesn't support the {type(value)} type.")



from locust import events

from connect.client.contrib.locust.user import ConnectHttpUser  # noqa


@events.init_command_line_parser.add_listener
def _(parser):
    parser.add_argument('--connect-api-key', default='', help='Connect Api Key')



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from connect.client.rql.utils import parse_kwargs, to_rql_value


class RQLQuery:
    """
    Helper class to construct complex RQL queries.
    """
    AND = 'and'
    OR = 'or'
    EXPR = 'expr'

    def __init__(self, *, _op=EXPR, _children=None, _negated=False, _expr=None, **kwargs):
        """
        Create a new R object.

        Query filters can be expressed through keyword arguments or using methods.

        Ex.

        .. code-block:: python

            rql = R(field='value', field2__in=('v1', 'v2'), field3__empty=True)

        All the lookups expressed as keyword arguments are combined together with a logical ``and``.

        Using the ``n`` method:

        .. code-block:: python

            rql = (
                R().n('field').eq('value')
                & R().n('field2').in_(('v1', 'v2'))
                & R().n('field3').empty(True)
            )

        The previous query can be expressed in a more concise form like:

        .. code-block:: python

            rql = R().field.eq('value') & R().field2.in_(('v1', 'v2')) & r.field3.empty(True)

        The R object support the bitwise operators ``&``, ``|`` and ``~``.

        Nested fields can be expressed using dot notation:

        .. code-block:: python

            rql = R().n('nested.field').eq('value')

        or

        .. code-block:: python

            rql = R().nested.field.eq('value')
        """
        self.op = _op
        self.children = _children or []
        self.negated = _negated
        self.expr = _expr
        self._path = []
        self._field = None
        if len(kwargs) == 1:
            self.op = self.EXPR
            self.expr = parse_kwargs(kwargs)[0]
        if len(kwargs) > 1:
            self.op = self.AND
            for token in parse_kwargs(kwargs):
                self.children.append(RQLQuery(_expr=token))

    def __len__(self):
        """
        Returns the length of this R object.
        It will be 1 if represent a single RQL expression
        or the number of expressions this logical operator (and, or)
        applies to.

        :return: The length of this R object.
        :rtype: int
        """
        if self.op == self.EXPR:
            if self.expr:
                return 1
            return 0
        return len(self.children)

    def __bool__(self):
        """
        Returns False if it is an empty R object otherwise True.

        :return: False if it is an empty R object otherwise True.
        :rtype: bool
        """
        return bool(self.children) or bool(self.expr)

    def __eq__(self, other):
        """
        Returns True if self == other.

        :param other: Another R object.
        :type other: R
        :return: True if the ``other`` object is equal to self, False otherwise.
        :rtype: bool
        """
        return (
            self.op == other.op
            and self.children == other.children
            and self.negated == other.negated
            and self.expr == other.expr
        )

    def __hash__(self):
        """
        Calculate a hash that identifies uniquely this R object.

        :return: The hash of this object.
        :rtype: int
        """
        return hash(
            (self.op, self.expr, self.negated, *(hash(value) for value in self.children)),
        )

    def __repr__(self):
        if self.op == self.EXPR:
            return f'<R({self.op}) {self.expr}>'
        return f'<R({self.op})>'

    def __and__(self, other):
        """
        Combine this R object with ``other`` using a logical ``and``.

        :param other: Another R object.
        :type other: R
        :return: The R object representing a logical ``and`` between this and ``other``.
        :rtype: R
        """
        return self._join(other, self.AND)

    def __or__(self, other):
        """
        Combine this R object with ``other`` using a logical ``or``.

        :param other: Another R object.
        :type other: R
        :return: The R object representing a logical ``or`` between this and ``other``.
        :rtype: R
        """
        return self._join(other, self.OR)

    def __invert__(self):
        """
        Apply the RQL ``not`` operator to this R object.

        :return: The R object representing this R object negated.
        :rtype: R
        """
        query = RQLQuery(_op=self.AND, _expr=self.expr, _negated=True)
        query._append(self)
        return query

    def __getattr__(self, name):
        return self.n(name)

    def __str__(self):
        return self._to_string(self)

    def n(self, name):
        """
        Set the current field for this R object.

        :param name: name of the field
        :type name: str
        :raises AttributeError: if this R object has already been evaluated.
        :return: This R object.
        :rtype: R
        """
        if self._field:
            raise AttributeError('Already evaluated')

        self._path.extend(name.split('.'))
        return self

    def ne(self, value):
        return self._bin('ne', value)

    def eq(self, value):
        return self._bin('eq', value)

    def lt(self, value):
        return self._bin('lt', value)

    def le(self, value):
        return self._bin('le', value)

    def gt(self, value):
        return self._bin('gt', value)

    def ge(self, value):
        return self._bin('ge', value)

    def out(self, value):
        return self._list('out', value)

    def in_(self, value):
        return self._list('in', value)

    def oneof(self, value):
        return self._list('in', value)

    def null(self, value):
        return self._bool('null', value)

    def empty(self, value):
        return self._bool('empty', value)

    def like(self, value):
        return self._bin('like', value)

    def ilike(self, value):
        return self._bin('ilike', value)

    def _bin(self, op, value):
        self._field = '.'.join(self._path)
        value = to_rql_value(op, value)
        self.expr = f'{op}({self._field},{value})'
        return self

    def _list(self, op, value):
        self._field = '.'.join(self._path)
        value = to_rql_value(op, value)
        self.expr = f'{op}({self._field},({value}))'
        return self

    def _bool(self, expr, value):
        self._field = '.'.join(self._path)
        if bool(value) is False:
            self.expr = f'ne({self._field},{expr}())'
            return self
        self.expr = f'eq({self._field},{expr}())'
        return self

    def _to_string(self, query):
        tokens = []
        if query.expr:
            if query.negated:
                return f'not({query.expr})'
            return query.expr
        for c in query.children:
            if c.expr:
                if c.negated:
                    tokens.append(f'not({c.expr})')
                else:
                    tokens.append(c.expr)
                continue
            tokens.append(self._to_string(c))

        if not tokens:
            return ''

        if query.negated:
            return f'not({query.op}({",".join(tokens)}))'
        return f'{query.op}({",".join(tokens)})'

    def _copy(self, other):
        return RQLQuery(
            _op=other.op,
            _children=other.children[:],
            _expr=other.expr,
        )

    def _join(self, other, op):
        if self == other:
            return self._copy(self)
        if not other:
            return self._copy(self)
        if not self:
            return self._copy(other)

        query = RQLQuery(_op=op)
        query._append(self)
        query._append(other)
        return query

    def _append(self, other):
        if other in self.children:
            return other

        if (
            (other.op == self.op or (len(other) == 1 and other.op != self.EXPR))
            and not other.negated
        ):
            self.children.extend(other.children)
            return self

        self.children.append(other)
        return self


R = RQLQuery



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
import inflect

from cmr import render

from connect.client.models import Action, Collection, NS, Resource, ResourceSet


_COL_HTTP_METHOD_TO_METHOD = {
    'get': '.all(), .filter(), .first(), .last()',
    'post': '.create()',
}


class DefaultFormatter:

    def __init__(self, specs):
        self._specs = specs
        self._p = inflect.engine()

    def format_client(self):
        lines = [
            f'# Welcome to {self._specs.title} {self._specs.version}',
            '## Introduction',
        ] + self._specs.description.splitlines()

        lines += [
            '',
            '## Namespaces',
        ]
        for ns in self._specs.get_namespaces():
            lines.append(f'* {ns}')

        lines += [
            '',
            '## Collections',
        ]
        for col in self._specs.get_collections():
            lines.append(f'* {col}')

        return render('\n'.join(lines))

    def format_ns(self, ns):
        namespaces = self._specs.get_nested_namespaces(ns.path)
        collections = self._specs.get_namespaced_collections(ns.path)

        if not (collections or namespaces):
            return render(f'~~{ns.path}~~ **does not exists.**')

        lines = [
            f'# {ns.path.title()} namespace',
            f'**path: /{ns.path}**',
        ]
        if namespaces:
            lines.append('## Available namespaces')
            for namespace in namespaces:
                lines.append(f'* {namespace}')

        if collections:
            lines.append('## Available collections')
            for collection in collections:
                lines.append(f'* {collection}')

        return render('\n'.join(lines))

    def format_collection(self, collection):
        col_info = self._specs.get_collection(collection.path)
        if not col_info:
            return render(f'~~{collection.path}~~ **does not exists.**')

        if '/' in collection.path:
            _, collection_name = collection.path.rsplit('/', 1)
        else:
            collection_name = collection.path

        lines = [
            f'# {collection_name.title()} collection',
            f'**path: /{collection.path}**',
        ]

        lines.extend(self._format_heading(col_info))
        lines.append('### Available operations')
        if 'get' in col_info:
            lines.append(f'* GET: {_COL_HTTP_METHOD_TO_METHOD["get"]}')

        if 'post' in col_info:
            lines.append(f'* POST: {_COL_HTTP_METHOD_TO_METHOD["post"]}')

        return render('\n'.join(lines))

    def format_resource(self, resource):
        res_info = self._specs.get_resource(resource.path)
        if not res_info:
            return render(f'~~{resource.path}~~ **does not exists.**')

        resource_name = resource.path.split('/')[-2]
        resource_name = self._p.singular_noun(resource_name)
        lines = [
            f'# {resource_name.title()} resource',
            f'**path: /{resource.path}**',
        ]

        lines.extend(self._format_heading(res_info))

        actions = self._specs.get_actions(resource.path)
        if actions:
            lines.append('### Available actions')
            for name, summary in actions:
                if summary:
                    lines.append(f'* {name} - {summary}')
                else:
                    lines.append(f'* {name}')
            lines.append('')

        nested = self._specs.get_nested_collections(resource.path)
        if nested:
            lines.append('### Available nested collections')
            for name, summary in nested:
                if summary:
                    lines.append(f'* {name} - {summary}')
                else:
                    lines.append(f'* {name}')

        return render('\n'.join(lines))

    def format_action(self, action):
        action_info = self._specs.get_action(action.path)
        if not action_info:
            return render(f'~~{action.path}~~ **does not exists.**')

        _, action_name = action.path.rsplit('/', 1)

        lines = [
            f'# {action_name.title()} action',
            f'**path: /{action.path}**',
        ]
        lines.extend(self._format_heading(action_info))
        lines.append('## Available methods')
        for method in ('get', 'post', 'put', 'delete'):
            if method in action_info:
                lines.append(f'* {method.upper()}: .{method}()')

        return render('\n'.join(lines))

    def format_resource_set(self, rs):
        col_info = self._specs.get_collection(rs.path)
        if not col_info:
            return render(f'~~{rs.path}~~ **does not exists.**')

        if '/' in rs.path:
            _, collection_name = rs.path.rsplit('/', 1)
        else:
            collection_name = rs.path

        lines = [
            f'# Search the {collection_name.title()} collection',
            f'**path: /{rs.path}**',
        ]
        lines.extend(self._format_heading(col_info))
        get_info = col_info['get']
        params = get_info.get('parameters')
        if not params:
            return render('\n'.join(lines))
        filters = list(sorted(filter(lambda x: '$ref' not in x, params), key=lambda x: x['name']))
        lines.append(f'Support pagination: *{len(params) > len(filters)}*')
        if not filters:
            return render('\n'.join(lines))

        lines.append('## Available filters')
        for filter_ in filters:
            lines.append(f'*{filter_["name"]}*')
            description = filter_['description'].splitlines()
            for line in description:
                line = line.strip()
                if line:
                    lines.append(f'   {line}')
            lines.append('')

        return render('\n'.join(lines))

    def format(self, obj):
        if not self._specs:
            return render('**No OpenAPI specs available.**')

        if isinstance(obj, NS):
            return self.format_ns(obj)

        if isinstance(obj, Collection):
            return self.format_collection(obj)

        if isinstance(obj, Resource):
            return self.format_resource(obj)

        if isinstance(obj, Action):
            return self.format_action(obj)

        if isinstance(obj, ResourceSet):
            return self.format_resource_set(obj)

        return self.format_client()

    def _format_heading(self, info):
        lines = []
        for section in ('summary', 'description'):
            if section in info:
                lines.append(f'### {section.title()}')
                lines.append(info[section])
        return lines



import json
import sys


class RequestLogger:
    def __init__(self, file=sys.stdout):
        self._file = file

    def obfuscate(self, value):
        if value.startswith('ApiKey SU-'):
            return value.split(':')[0] + '*' * 10
        else:
            return '*' * 20

    def log_request(self, method, url, kwargs):
        other_args = {k: v for k, v in kwargs.items() if k not in ('headers', 'json', 'params')}

        if 'params' in kwargs:
            url += '&' if '?' in url else '?'
            url += '&'.join([f'{k}={v}' for k, v in kwargs['params'].items()])

        lines = [
            '--- HTTP Request ---',
            f'{method.upper()} {url} {other_args if other_args else ""}',
        ]

        if 'headers' in kwargs:
            for k, v in kwargs['headers'].items():
                if k == 'Authorization':
                    v = self.obfuscate(v)
                lines.append(f'{k}: {v}')

        if 'json' in kwargs:
            lines.append(json.dumps(kwargs['json'], indent=4))

        lines.append('')

        print(*lines, sep='\n', file=self._file)

    def log_response(self, response):
        reason = response.raw.reason if getattr(response, 'raw', None) else response.reason_phrase
        lines = [
            '--- HTTP Response ---',
            f'{response.status_code} {reason}',
        ]

        for k, v in response.headers.items():
            lines.append(f'{k}: {v}')

        if response.headers.get('Content-Type', None) == 'application/json':
            lines.append(json.dumps(response.json(), indent=4))

        lines.append('')

        print(*lines, sep='\n', file=self._file)



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
CONNECT_SPECS_URL = 'https://apispec.connect.cloudblue.com/connect-openapi30.yml'  # noqa
CONNECT_ENDPOINT_URL = 'https://api.connect.cloudblue.com/public/v1'  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cnct',
 'connect',
 'connect.client',
 'connect.client.contrib',
 'connect.client.contrib.locust',
 'connect.client.models',
 'connect.client.rql']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1',
 'asgiref>=3.3.4,<4.0.0',
 'connect-markdown-renderer>=1.0.0,<2.0.0',
 'httpx>=0.18.1,<0.19.0',
 'inflect>=4.1',
 'requests>=2.23']

setup_kwargs = {
    'name': 'connect-openapi-client',
    'version': '24.5',
    'description': 'Connect Python OpenAPI Client',
    'long_description': '# Connect Python OpenAPI Client\n\n![pyversions](https://img.shields.io/pypi/pyversions/connect-openapi-client.svg) [![PyPi Status](https://img.shields.io/pypi/v/connect-openapi-client.svg)](https://pypi.org/project/connect-openapi-client/) [![Build Status](https://github.com/cloudblue/connect-python-openapi-client/workflows/Build%20Connect%20Python%20OpenAPI%20Client/badge.svg)](https://github.com/cloudblue/connect-python-openapi-client/actions) [![codecov](https://codecov.io/gh/cloudblue/connect-python-openapi-client/branch/master/graph/badge.svg)](https://codecov.io/gh/cloudblue/connect-python-openapi-client) [![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=connect-open-api-client&metric=alert_status)](https://sonarcloud.io/dashboard?id=connect-open-api-client)\n\n\n\n\n## Introduction\n\n`Connect Python OpenAPI Client` is the simple, concise, powerful and REPL-friendly CloudBlue Connect API client.\n\nIt has been designed following the [fluent interface design pattern](https://en.wikipedia.org/wiki/Fluent_interface).\n\nDue to its REPL-friendly nature, using the CloudBlue Connect OpenAPI specifications it allows developers to learn and\nplay with the CloudBlue Connect API using a python REPL like [jupyter](https://jupyter.org/) or [ipython](https://ipython.org/).\n\n\n## Install\n\n`Connect Python OpenAPI Client` requires python 3.6 or later and has the following dependencies:\n\n* connect-markdown-renderer>=1,<2\n* pyyaml>=5,<6\n* requests>=2,<3\n\n`Connect Python OpenAPI Client` can be installed from [pypi.org](https://pypi.org/project/connect-openapi-client/) using pip:\n\n```\n$ pip install connect-openapi-client\n```\n\n\n## Documentation\n\n[`Connect Python OpenAPI Client` documentation](https://connect-openapi-client.readthedocs.io/en/latest/) is hosted on the _Read the Docs_ service.\n\n\n## License\n\n``Connect Python OpenAPI Client`` is released under the [Apache License Version 2.0](https://www.apache.org/licenses/LICENSE-2.0).\n',
    'author': 'CloudBlue',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from connect.client.models.base import (  # noqa
    Action,
    AsyncAction,
    AsyncCollection,
    AsyncNS,
    AsyncResource,
    Collection,
    NS,
    Resource,
)
from connect.client.models.exceptions import NotYetEvaluatedError  # noqa
from connect.client.models.resourceset import AsyncResourceSet, ResourceSet  # noqa


#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from connect.client.exceptions import ClientError
from connect.client.utils import resolve_attribute


class CollectionMixin:
    def create(self, payload=None, **kwargs):
        """
        Create a new resource within this collection.

        :param payload: JSON payload of the resource to create, defaults to None.
        :type payload: dict, optional
        :return: The newly created resource.
        :rtype: dict
        """
        return self._client.create(
            self._path,
            payload=payload,
            **kwargs,
        )


class AsyncCollectionMixin:
    async def create(self, payload=None, **kwargs):
        """
        Create a new resource within this collection.

        :param payload: JSON payload of the resource to create, defaults to None.
        :type payload: dict, optional
        :return: The newly created resource.
        :rtype: dict
        """
        return await self._client.create(
            self._path,
            payload=payload,
            **kwargs,
        )


class ResourceMixin:
    def exists(self):
        try:
            self.get()
            return True
        except ClientError as ce:
            if ce.status_code and ce.status_code == 404:
                return False
            raise

    def get(self, **kwargs):
        """
        Execute a http GET to retrieve this resource.
        The http GET can be customized passing kwargs that
        will be forwarded to the underlying GET of the ``requests``
        library.

        :return: The resource data.
        :rtype: dict
        """
        return self._client.get(self._path, **kwargs)

    def update(self, payload=None, **kwargs):
        """
        Execute a http PUT to update this resource.
        The http PUT can be customized passing kwargs that
        will be forwarded to the underlying PUT of the ``requests``
        library.

        :param payload: the JSON payload of the update request, defaults to None
        :type payload: dict, optional
        :return: The updated resource.
        :rtype: dict
        """
        return self._client.update(
            self._path,
            payload=payload,
            **kwargs,
        )

    def delete(self, **kwargs):
        """
        Execute a http DELETE to delete this resource.
        The http DELETE can be customized passing kwargs that
        will be forwarded to the underlying DELETE of the ``requests``
        library.
        """
        return self._client.delete(
            self._path,
            **kwargs,
        )

    def values(self, *fields):
        """
        Returns a flat dictionary containing only the fields passed as arguments.
        Nested field can be specified using dot notation.

        Ex.

        .. code-block:: python

        values = resource.values('field', 'nested.field')

        :return: A dictionary containing field,value pairs.
        :rtype: dict
        """
        results = {}
        item = self.get()
        for field in fields:
            results[field] = resolve_attribute(field, item)
        return results


class AsyncResourceMixin:
    async def exists(self):
        try:
            await self.get()
            return True
        except ClientError as ce:
            if ce.status_code and ce.status_code == 404:
                return False
            raise

    async def get(self, **kwargs):
        """
        Execute a http GET to retrieve this resource.
        The http GET can be customized passing kwargs that
        will be forwarded to the underlying GET of the ``requests``
        library.

        :return: The resource data.
        :rtype: dict
        """
        return await self._client.get(self._path, **kwargs)

    async def update(self, payload=None, **kwargs):
        """
        Execute a http PUT to update this resource.
        The http PUT can be customized passing kwargs that
        will be forwarded to the underlying PUT of the ``requests``
        library.

        :param payload: the JSON payload of the update request, defaults to None
        :type payload: dict, optional
        :return: The updated resource.
        :rtype: dict
        """
        return await self._client.update(
            self._path,
            payload=payload,
            **kwargs,
        )

    async def delete(self, **kwargs):
        """
        Execute a http DELETE to delete this resource.
        The http DELETE can be customized passing kwargs that
        will be forwarded to the underlying DELETE of the ``requests``
        library.
        """
        return await self._client.delete(
            self._path,
            **kwargs,
        )

    async def values(self, *fields):
        """
        Returns a flat dictionary containing only the fields passed as arguments.
        Nested field can be specified using dot notation.

        Ex.

        .. code-block:: python

        values = resource.values('field', 'nested.field')

        :return: A dictionary containing field,value pairs.
        :rtype: dict
        """
        results = {}
        item = await self.get()
        for field in fields:
            results[field] = resolve_attribute(field, item)
        return results


class ActionMixin:
    def get(self, **kwargs):
        """
        Execute this action through a http GET.
        The http GET can be customized passing kwargs that
        will be forwarded to the underlying GET of the ``requests``
        library.

        :return: The action data.
        :rtype: dict, None
        """
        return self._client.get(self._path, **kwargs)

    def post(self, payload=None, **kwargs):
        """
        Execute this action through a http POST.
        The http POST can be customized passing kwargs that
        will be forwarded to the underlying PUT of the ``requests``
        library.

        :param payload: the JSON payload for this action, defaults to None
        :type payload: dict, optional
        :return: The result of this action.
        :rtype: dict, None
        """
        if payload:
            kwargs['json'] = payload
        return self._client.execute(
            'post',
            self._path,
            **kwargs,
        )

    def put(self, payload=None, **kwargs):
        """
        Execute this action through a http PUT.
        The http PUT can be customized passing kwargs that
        will be forwarded to the underlying PUT of the ``requests``
        library.

        :param payload: the JSON payload for this action, defaults to None
        :type payload: dict, optional
        :return: The result of this action.
        :rtype: dict, None
        """
        if payload:
            kwargs['json'] = payload
        return self._client.execute(
            'put',
            self._path,
            **kwargs,
        )

    def delete(self, **kwargs):
        """
        Execute this action through a http DELETE.
        The http DELETE can be customized passing kwargs that
        will be forwarded to the underlying DELETE of the ``requests``
        library.
        """
        return self._client.delete(
            self._path,
            **kwargs,
        )


class AsyncActionMixin:
    async def get(self, **kwargs):
        """
        Execute this action through a http GET.
        The http GET can be customized passing kwargs that
        will be forwarded to the underlying GET of the ``requests``
        library.

        :return: The action data.
        :rtype: dict, None
        """
        return await self._client.get(self._path, **kwargs)

    async def post(self, payload=None, **kwargs):
        """
        Execute this action through a http POST.
        The http POST can be customized passing kwargs that
        will be forwarded to the underlying PUT of the ``requests``
        library.

        :param payload: the JSON payload for this action, defaults to None
        :type payload: dict, optional
        :return: The result of this action.
        :rtype: dict, None
        """
        if payload:
            kwargs['json'] = payload
        return await self._client.execute(
            'post',
            self._path,
            **kwargs,
        )

    async def put(self, payload=None, **kwargs):
        """
        Execute this action through a http PUT.
        The http PUT can be customized passing kwargs that
        will be forwarded to the underlying PUT of the ``requests``
        library.

        :param payload: the JSON payload for this action, defaults to None
        :type payload: dict, optional
        :return: The result of this action.
        :rtype: dict, None
        """
        if payload:
            kwargs['json'] = payload
        return await self._client.execute(
            'put',
            self._path,
            **kwargs,
        )

    async def delete(self, **kwargs):
        """
        Execute this action through a http DELETE.
        The http DELETE can be customized passing kwargs that
        will be forwarded to the underlying DELETE of the ``requests``
        library.
        """
        return await self._client.delete(
            self._path,
            **kwargs,
        )



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from http import HTTPStatus


class ClientError(Exception):
    def __init__(self, message=None, status_code=None, error_code=None, errors=None, **kwargs):
        self.message = message
        self.status_code = status_code
        self.error_code = error_code
        self.errors = errors
        self.additional_info = kwargs

    def __repr__(self):
        return f'<ClientError {self.status_code}: {self.error_code}>'

    def __str__(self):
        message = self.message or self._get_status_description() or 'Unexpected error'
        if self.error_code and self.errors:
            errors = ','.join(self.errors)
            return f'{message}: {self.error_code} - {errors}'
        return message

    def _get_status_description(self):
        if not self.status_code:
            return
        status = HTTPStatus(self.status_code)
        description = status.name.replace('_', ' ').title()
        return f'{self.status_code} {description}'



#
# This file is part of the Ingram Micro CloudBlue Connect Python OpenAPI Client.
#
# Copyright (c) 2021 Ingram Micro. All Rights Reserved.
#
from connect.client.utils import get_values, parse_content_range


class aiter:
    def __init__(self, values):
        self._values = iter(values)

    async def __anext__(self):
        try:
            return next(self._values)
        except StopIteration:
            raise StopAsyncIteration


class AbstractBaseIterator:
    def __init__(self, rs, client, path, query, config, **kwargs):
        self._rs = rs
        self._results_iterator = None
        self._client = client
        self._path = path
        self._query = query
        self._config = config
        self._kwargs = kwargs
        self._loaded = False

    def get_item(self, item):
        raise NotImplementedError('get_item must be implemented in subclasses.')


class AbstractIterator(AbstractBaseIterator):

    def _load(self):
        if not self._loaded:
            self._rs._results, self._rs._content_range = self._execute_request()
            self._results_iterator = iter(self._rs._results)
            self._loaded = True

    def __next__(self):
        self._load()

        if not self._rs._results:
            raise StopIteration
        try:
            item = next(self._results_iterator)
        except StopIteration:
            if (
                self._rs._content_range is None
                or self._rs._content_range.last >= self._rs._content_range.count - 1
            ):
                raise
            self._config['params']['offset'] += self._config['params']['limit']
            results, cr = self._execute_request()
            if not results:
                raise
            self._rs._content_range = cr
            self._rs._results.extend(results)
            self._results_iterator = iter(results)
            item = next(self._results_iterator)

        return self.get_item(item)

    def _execute_request(self):
        results = self._client.get(
            f'{self._path}?{self._query}',
            **self._config,
        )
        content_range = parse_content_range(
            self._client.response.headers.get('Content-Range'),
        )
        return results, content_range


class AbstractAsyncIterator(AbstractBaseIterator):

    async def _load(self):
        if not self._loaded:
            self._rs._results, self._rs._content_range = await self._execute_request()
            self._results_iterator = iter(self._rs._results)
            self._loaded = True

    async def __anext__(self):
        await self._load()

        if not self._rs._results:
            raise StopAsyncIteration
        try:
            item = next(self._results_iterator)
        except StopIteration:
            if (
                self._rs._content_range is None
                or self._rs._content_range.last >= self._rs._content_range.count - 1
            ):
                raise StopAsyncIteration
            self._config['params']['offset'] += self._config['params']['limit']
            results, cr = await self._execute_request()
            if not results:
                raise StopAsyncIteration
            self._rs._content_range = cr
            self._rs._results.extend(results)
            self._results_iterator = iter(results)
            item = next(self._results_iterator)

        return self.get_item(item)

    async def _execute_request(self):
        results = await self._client.get(
            f'{self._path}?{self._query}',
            **self._config,
        )
        content_range = parse_content_range(
            self._client.response.headers.get('Content-Range'),
        )
        return results, content_range


class ResourceMixin:
    def get_item(self, item):
        return item


class ValueListMixin:
    def get_item(self, item):
        return get_values(item, self._kwargs['fields'])


class ResourceIterator(ResourceMixin, AbstractIterator):
    pass


class ValuesListIterator(ValueListMixin, AbstractIterator):
    pass


class AsyncResourceIterator(ResourceMixin, AbstractAsyncIterator):
    pass


class AsyncValuesListIterator(ValueListMixin, AbstractAsyncIterator):
    pass
